import torch
import torch.nn as nn
from torch.distributions import Dirichlet
import numpy as np
from scipy.special import multigammaln
from model.mvn import MVN
from model.prior import Prior
from model.utils import cov
from model.gamma import Gamma
from model.logdet import logdet

import time

class GMM(nn.Module):
    def __init__(self, dimension, num_components, device):
        super(GMM, self).__init__()
        self.dimension = torch.tensor(dimension)
        self.num_components = torch.tensor(num_components)
        self.U = nn.Parameter(torch.randn([dimension, dimension], dtype=torch.float32, device=device)
                              + torch.eye(dimension, dtype=torch.float32, device=device))
        self.p = nn.Parameter(torch.abs(torch.ones([num_components], dtype=torch.float32, device=device)/num_components))
        self.D = [MVN(dimension, device) for i in range(self.num_components)]
        self.prior = Prior()
        self.device = device

    def initialize(self, data_matrix, device):
        data = data_matrix
        num_data = len(data)
        data = data[torch.randperm(num_data), :]
        if device.type == 'cuda':
            self.p.data = torch.ones(self.num_components).cuda() / self.num_components
        else:
            self.p.data = torch.ones(self.num_components) / self.num_components
        sigma = cov(data)
        _, self.U.data = torch.eig(sigma, eigenvectors=True)
        label = self.num_components * torch.rand(num_data, device=device)
        label = torch.ceil(label)
        for k in range(self.num_components):
            indices_k = np.where(label.cpu() == k + 1)[0]
            data_k = data[indices_k, :]
            mean_k = torch.mean(data_k, 0)
            self.D[k].mu.data = mean_k
            sigma_k = cov(data_k)
            temp, _ = torch.eig(sigma_k, eigenvectors=False)
            self.D[k].d.data = torch.log(temp[:, 0])
        self.prior = Prior(data, self.num_components)


    def forward(self, x, ndata):
        start = time.time()
        m = nn.Softmax()
        p_softmax = m(self.p)
        if torch.cuda.is_available():
            hx0 = torch.zeros([self.num_components, x.shape[0]]).cuda()
        else:
            hx0 = torch.zeros([self.num_components, x.shape[0]])
        for k in range(self.num_components):
            ll_vec_k = self.D[k].ll(self.U, x)
            hx0[k, :] = ll_vec_k + torch.log(p_softmax[k])
        hx = hx0.clone()
        ll_vec = torch.logsumexp(hx, 0)
        start = time.time()
        log_p = self.log_prior()
        ll_vec = ll_vec #+ log_p[0]/ndata
        end = time.time()
        return ll_vec



    def nllloss(self, ll_vec, ndata):
        start = time.time()
        ll = torch.mean(ll_vec)
        end = time.time()
        return -ll

    def log_wishart(self, x):
        n = self.prior.nup
        v = self.prior.Sp
        p = x.shape[0]
        sld = torch.slogdet(x)
        log = 0
        log += (n-p-1)/2.0 * sld[1]


        log -= 0.5*torch.trace(torch.matmul(v, x))
        log -= (n*p)/2.0*np.log(2.0)
        sld = torch.slogdet(v)
        log -= n/2.0*sld[1]


        log -= multigammaln(n/2.0, p)
        return log

    def log_prior(self):
        start = time.time()
        lp_uu = 0
        for k in range(self.num_components):
            dk = self.D[k].d
            Dk = torch.diag(dk)
            x = torch.matmul(self.U, Dk)
            x = torch.matmul(x, torch.transpose(self.U, 0, 1))
            lp_uu += self.log_wishart(x)

        m = nn.Softmax()
        x = m(self.p)
        if self.device.type == 'cuda':
            a = 1e-2 * torch.ones(self.num_components).cuda()
        else:
            a = 1e-2 * torch.ones(self.num_components)
        dirichlet = Dirichlet(a)
        lp_p = dirichlet.log_prob(x)
        lp_p = torch.sum(lp_p)

        if self.device.type == 'cuda':
            lp_mu = torch.zeros(1).cuda()
        else:
            lp_mu = torch.zeros(1)
        for k in range(self.num_components):
            mu = self.D[k].mu
            dk = self.D[k].d
            Dk = torch.diag(dk)
            if self.device.type == 'cuda':
                y = torch.zeros(1).cuda()
            else:
                y = torch.zeros(1)
            y += -self.dimension*np.log(2*np.pi)
            y += self.dimension*np.log(self.prior.kp)
            sld = torch.slogdet(self.U)
            y += 2*sld[1]

            y += torch.sum(torch.log(dk))
            y *= 0.5
            s = torch.matmul(torch.transpose(self.U, 0, 1), mu - self.prior.mup)
            s = torch.sqrt(dk) * s
            s = -self.prior.kp/2*torch.dot(s, s)
            y += s
            lp_mu += y
        end = time.time()

        return lp_mu + lp_p + lp_uu

