import os
import numpy as np
import scipy as sp
import scipy.io as sio
import torch


def cov(m, rowvar=False):
    '''Estimate a covariance matrix given data.

    Covariance indicates the level to which two variables vary together.
    If we examine N-dimensional samples, `X = [x_1, x_2, ... x_N]^T`,
    then the covariance matrix element `C_{ij}` is the covariance of
    `x_i` and `x_j`. The element `C_{ii}` is the variance of `x_i`.

    Args:
        m: A 1-D or 2-D array containing multiple variables and observations.
            Each row of `m` represents a variable, and each column a single
            observation of all those variables.
        rowvar: If `rowvar` is True, then each row represents a
            variable, with observations in the columns. Otherwise, the
            relationship is transposed: each column represents a variable,
            while the rows contain observations.

    Returns:
        The covariance matrix of the variables.
    '''
    if m.dim() > 2:
        raise ValueError('m has more than 2 dimensions')
    if m.dim() < 2:
        m = m.view(1, -1)
    if not rowvar and m.size(0) != 1:
        m = m.t()
    fact = 1.0 / (m.size(1) - 1)
    m -= torch.mean(m, dim=1, keepdim=True)
    mt = m.t()  # if complex: mt = m.t().conj()
    return fact * m.matmul(mt).squeeze()


def write_log(log_path, message):
    log_file = open(log_path, 'a+')
    log_file.write(message)
    log_file.write('\n')
    log_file.close()


def load_data(data_directory, data_set, mat_name):
    data_set_mat = data_set + '.mat'
    data_set_mat = os.path.join(data_directory, data_set_mat)
    mat = sio.loadmat(data_set_mat)
    mdata = mat[mat_name]  # variable in mat file

    mtrain = mdata['train']  # variable in mat file
    mtrain = mtrain[0, 0]
    features = mtrain
    features = features.astype('float64')
    sigma = np.cov(features)
    features = features.T
    mean = np.mean(features, axis=0)
    features -= mean
    features = np.matmul(features, np.linalg.inv(sp.linalg.sqrtm(sigma)))

    mtest = mdata['test']  # variable in mat file
    mtest = mtest[0, 0]
    tfeatures = mtest
    tfeatures = tfeatures.astype('float64')
    tfeatures = tfeatures.T
    tfeatures -= mean
    tfeatures = np.matmul(tfeatures, np.linalg.inv(sp.linalg.sqrtm(sigma)))
    return features, tfeatures


def get_lr(optimizer):
    for param_group in optimizer.param_groups:
        return param_group['lr']


def str2bool(v):
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

