import os
import sys
import time
import argparse
from datetime import datetime
import numpy as np
import scipy as sp
import scipy.io as sio
from torch.autograd import Variable
import torch
torch.manual_seed(0)
import torch.utils.data as data_utils
import torch.optim as optim
from model.gmm import GMM
from model.utils import write_log, load_data, get_lr, str2bool

# hyper parameters
batch_size = 128
display_iteration = 10
epoch_no = 300
decay_steps = 25


def train_model(train_features, test_features, model_path, log_file_name, optimizer_name, lr_scheduler=False, use_gpu=False, restore_path=None):
    device = torch.device('cuda' if torch.cuda.is_available() and use_gpu else 'cpu')
    since = time.time()
    features = torch.from_numpy(train_features)
    tfeatures = torch.from_numpy(test_features)
    if device.type == 'cuda':
        features = features.type(torch.cuda.FloatTensor)
        tfeatures = tfeatures.type(torch.cuda.FloatTensor)
    else:
        features = features.type(torch.FloatTensor)
        tfeatures = tfeatures.type(torch.FloatTensor)
    train_set = data_utils.TensorDataset(features)
    test_set = data_utils.TensorDataset(tfeatures)
    dataset = {'train': train_set, 'test': test_set}
    dataset_sizes = {'train': features.shape[0], 'test': tfeatures.shape[0]}
    train_loader = data_utils.DataLoader(train_set, batch_size=batch_size, shuffle=True)
    test_loader = data_utils.DataLoader(test_set, batch_size=batch_size, shuffle=True)
    dataloaders = {'train': train_loader, 'test': test_loader}
    ndata, dimension = features.shape
    gmm = GMM(dimension, dimension, device)
    if torch.no_grad():
        gmm.initialize(features, device)
    
    print('use_gpu: {}'.format(use_gpu))
    if restore_path:
        gmm = torch.load(restore_path)
    optimizer_dict = {
        'sgd_lr_2': optim.SGD(gmm.parameters(), lr=1e-2),
        'sgd_lr_3': optim.SGD(gmm.parameters(), lr=1e-3),
        'sgd_lr_2_momentum': optim.SGD(gmm.parameters(), lr=1e-2, momentum=0.9),
        'sgd_lr_3_momentum': optim.SGD(gmm.parameters(), lr=1e-3, momentum=0.9),
        'sgd_lr_4_momentum': optim.SGD(gmm.parameters(), lr=1e-4, momentum=0.9),
        'adam_lr_2': optim.Adam(gmm.parameters(), lr=1e-2),
        'adam_lr_3': optim.Adam(gmm.parameters(), lr=1e-3),
        'adam_lr_4': optim.Adam(gmm.parameters(), lr=1e-4),
    }
    optimizer = optimizer_dict[optimizer_name]
    scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=decay_steps, gamma=0.5)
    ndata = features.shape[0]
    best_model_wts = gmm.state_dict()
    best_loss = sys.float_info.max
    best_grad = sys.float_info.max
    best_epoch = -1
    iteration = 0
    for epoch in range(epoch_no):  # loop over the dataset multiple times
        print('Epoch {}/{}'.format(epoch, epoch_no - 1))
        print('learning rate: {}'.format(get_lr(optimizer)))
        print('-' * 10)
        for phase in ['train', 'test']:
            if phase == 'train' and epoch % decay_steps == decay_steps - 1:
                scheduler.step()
            if phase == 'train':
                gmm.train(True)  # Set model to training mode
            else:
                gmm.train(False)  # Set model to evaluate mode
            running_loss = 0.0
            running_grad = 0.0
            epoch_iteration = 0
            epoch_loss = 0
            epoch_grad = 0
            for data in dataloaders[phase]:
                start0 = time.time()
                if phase == 'train':
                    iteration += 1
                                
                epoch_iteration += 1
                inputs = data[0]
                if use_gpu:
                    inputs = inputs.to(device) 

                # zero the parameter gradients
                optimizer.zero_grad()

                # forward pass
                ll_vec = gmm.forward(inputs, ndata)
                loss = gmm.nllloss(ll_vec, ndata)
                # backward + optimize only if in training phase
                start = time.time()
                loss.backward()
                end = time.time()

                start = time.time()
                if phase == 'train':
                    optimizer.step()
                end = time.time()

                running_loss += loss.item()
                running_grad += torch.norm(gmm.U.grad)
                epoch_loss += loss.item()
                epoch_grad += torch.norm(gmm.U.grad)
                if epoch_iteration % display_iteration == display_iteration-1:    # print every 1000 mini-batches
                    running_loss /= display_iteration
                    running_grad /= display_iteration
                    message = ('phase: %s, [%d, %5d] loss_batch: %.3f grad_batch: %.6f ' %
                               (phase, epoch + 1, iteration + 1, running_loss, running_grad))
                    print(message)
                    running_loss = 0.0
                    running_grad = 0.0
                end = time.time()

            epoch_loss /= epoch_iteration
            epoch_grad /= epoch_iteration
            message = ('Epoch:{}, Phase:{}, learning rate: {}, Loss: {}, Grad: {}'
                       .format(epoch, phase, get_lr(optimizer), epoch_loss, epoch_grad))
            print(message)
            write_log(log_file_name, message)
            checkpoint_path = "{}_{}_{}".format(model_path, epoch, iteration)
            if phase == 'train' and epoch%1 == 0:
                save_checkpoint(gmm, optimizer, get_lr(optimizer), epoch, iteration, checkpoint_path, lr_scheduler)

            if phase == 'test' and epoch_loss < best_loss:
                best_loss = epoch_loss
                best_grad = epoch_grad
                best_epoch = epoch
                best_model_wts = gmm.state_dict()
            
            
        # report the training phase
        time_elapsed = time.time() - since
        print('Training complete in {:.0f}m {:.0f}s'.format(
            time_elapsed // 60, time_elapsed % 60))
        message = ('Best Model Epoch:{}, Best Model Loss:{}, Best Model Grad:{}'
                   .format(best_epoch, best_loss, best_grad))
        write_log(log_file_name, message)

        # load best model weights
        gmm.load_state_dict(best_model_wts)
    return gmm


def save_checkpoint(model, optimizer, learning_rate, epoch, iteration, filepath, lr_scheduler):
    print("Saving model and optimizer state at iteration {} to {}".format(
          iteration, filepath))
    torch.save({'model': model,
                'epoch': epoch,
                'iteration': iteration,
                'lr_scheduler': lr_scheduler,
                'optimizer': optimizer.state_dict(),
                'learning_rate': learning_rate}, filepath+'.pt')
    torch.save({'model': model.state_dict(),
                'epoch': epoch,
                'iteration': iteration,
                'lr_scheduler': lr_scheduler,
                'optimizer': optimizer.state_dict(),
                'learning_rate': learning_rate}, filepath+'.pkl')


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--checkpoint_dir', type=str, required=True)
    parser.add_argument('--data_dir', type=str, required=True)
    parser.add_argument('--log_dir', type=str, required=True)
    parser.add_argument('--data_set', type=str, required=True)
    parser.add_argument('--data_name', type=str, required=True)
    parser.add_argument('--optimizer_name', type=str, required=True)
    parser.add_argument('--use_lr_scheduler', type=str2bool, required=True)
    parser.add_argument('--use_gpu', type=str2bool, required=True)
    parser.add_argument('--restore_path', type=str, required=False)

    args = parser.parse_args()
    features, tfeatures = load_data(data_directory=args.data_dir, data_set=args.data_set, mat_name=args.data_name)
    log_name = args.data_set + '_' + args.optimizer_name + '_' + str(args.use_lr_scheduler) + '.log'
    log_name = os.path.join(args.log_dir, log_name)

    model_file_name = args.data_set + '_' + args.optimizer_name + '_' + str(args.use_lr_scheduler) + '.pt'
    model_file_name = os.path.join(args.data_dir, model_file_name)

    print(args.use_lr_scheduler)
    checkpoint_path = 'checkpoint_' + args.data_set + '_' + args.optimizer_name + '_' + str(args.use_lr_scheduler) + '_' + str(batch_size)
    checkpoint_path = os.path.join(args.checkpoint_dir, checkpoint_path)

    if not os.path.exists(args.log_dir):
        os.makedirs(args.log_dir)

    if not os.path.exists(args.checkpoint_dir):
        os.makedirs(args.checkpoint_dir)

    model = train_model(train_features=features, test_features=tfeatures, log_file_name=log_name, model_path=checkpoint_path,
                        optimizer_name=args.optimizer_name, lr_scheduler=args.use_lr_scheduler, use_gpu=args.use_gpu, restore_path=args.restore_path)
    torch.save(model, model_file_name)


if __name__ == '__main__':
    main()
